@@ -18,7 +18,7 @@ import Data.Acquire (with)
 -- | For combinations of backends and entities that support
 -- cascade-deletion. “Cascade-deletion” means that entries that depend on
 -- other entries to be deleted will be deleted as well.
-class (PersistStore backend, PersistEntity record, backend ~ PersistEntityBackend record)
+class (PersistStoreWrite backend, PersistEntity record, BaseBackend backend ~ PersistEntityBackend record)
   => DeleteCascade record backend where
 
     -- | Perform cascade-deletion of single database
@@ -26,7 +26,7 @@ class (PersistStore backend, PersistEntity record, backend ~ PersistEntityBacken
     deleteCascade :: MonadIO m => Key record -> ReaderT backend m ()
 
 -- | Cascade-deletion of entries satisfying given filters.
-deleteCascadeWhere :: (MonadIO m, DeleteCascade record backend, PersistQuery backend)
+deleteCascadeWhere :: (MonadIO m, DeleteCascade record backend, PersistQueryWrite backend)
                    => [Filter record] -> ReaderT backend m ()
 deleteCascadeWhere filts = do
     srcRes <- selectKeysRes filts []
