@@ -1,8 +1,6 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE ExistentialQuantification #-}
-{-# LANGUAGE StandaloneDeriving #-}
 {-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE OverloadedStrings #-}
 #ifndef NO_OVERLAP
@@ -14,6 +12,7 @@ module Database.Persist.Class.PersistField
     , getPersistMap
     ) where
 
+import Control.Arrow (second)
 import Database.Persist.Types.Base
 import Data.Time (Day(..), TimeOfDay, UTCTime, parseTime)
 #ifdef HIGH_PRECISION_DATE
@@ -57,6 +56,8 @@ import System.Locale (defaultTimeLocale)
 
 #if MIN_VERSION_base(4,8,0)
 import Numeric.Natural (Natural)
+#else
+import Control.Applicative ((<$>))
 #endif
 
 -- | A value which can be marshalled to and from a 'PersistValue'.
@@ -176,28 +177,28 @@ instance PersistField Double where
     fromPersistValue x = Left $ T.pack $ "Expected Double, received: " ++ show x
 
 instance (HasResolution a) => PersistField (Fixed a) where
-  toPersistValue = PersistRational . toRational
-  fromPersistValue (PersistRational r) = Right $ fromRational r
-  fromPersistValue (PersistText t) = case reads $ T.unpack t of --  NOTE: Sqlite can store rationals just as string
-    [(a, "")] -> Right a
-    _ -> Left $ "Can not read " <> t <> " as Fixed"
-  fromPersistValue (PersistDouble d) = Right $ realToFrac d
-  fromPersistValue (PersistInt64 i) = Right $ fromIntegral i
-  fromPersistValue x = Left $ "PersistField Fixed:Expected Rational, received: " <> T.pack (show x)
+    toPersistValue = PersistRational . toRational
+    fromPersistValue (PersistRational r) = Right $ fromRational r
+    fromPersistValue (PersistText t) = case reads $ T.unpack t of --  NOTE: Sqlite can store rationals just as string
+      [(a, "")] -> Right a
+      _ -> Left $ "Can not read " <> t <> " as Fixed"
+    fromPersistValue (PersistDouble d) = Right $ realToFrac d
+    fromPersistValue (PersistInt64 i) = Right $ fromIntegral i
+    fromPersistValue x = Left $ "PersistField Fixed:Expected Rational, received: " <> T.pack (show x)
 
 instance PersistField Rational where
-  toPersistValue = PersistRational
-  fromPersistValue (PersistRational r) = Right r
-  fromPersistValue (PersistDouble d) = Right $ toRational d
-  fromPersistValue (PersistText t) = case reads $ T.unpack t of --  NOTE: Sqlite can store rationals just as string
-    [(a, "")] -> Right $ toRational (a :: Pico)
-    _ -> Left $ "Can not read " <> t <> " as Rational (Pico in fact)"
-  fromPersistValue (PersistInt64 i) = Right $ fromIntegral i
-  fromPersistValue (PersistByteString bs) = case double $ T.cons '0' $ T.decodeUtf8With T.lenientDecode bs of
-                                              Right (ret,"") -> Right $ toRational ret
-                                              Right (a,b) -> Left $ "Invalid bytestring[" <> T.pack (show bs) <> "]: expected a double but returned " <> T.pack (show (a,b))
-                                              Left xs -> Left $ "Invalid bytestring[" <> T.pack (show bs) <> "]: expected a double but returned " <> T.pack (show xs)
-  fromPersistValue x = Left $ "PersistField Rational:Expected Rational, received: " <> T.pack (show x)
+    toPersistValue = PersistRational
+    fromPersistValue (PersistRational r) = Right r
+    fromPersistValue (PersistDouble d) = Right $ toRational d
+    fromPersistValue (PersistText t) = case reads $ T.unpack t of --  NOTE: Sqlite can store rationals just as string
+      [(a, "")] -> Right $ toRational (a :: Pico)
+      _ -> Left $ "Can not read " <> t <> " as Rational (Pico in fact)"
+    fromPersistValue (PersistInt64 i) = Right $ fromIntegral i
+    fromPersistValue (PersistByteString bs) = case double $ T.cons '0' $ T.decodeUtf8With T.lenientDecode bs of
+                                                Right (ret,"") -> Right $ toRational ret
+                                                Right (a,b) -> Left $ "Invalid bytestring[" <> T.pack (show bs) <> "]: expected a double but returned " <> T.pack (show (a,b))
+                                                Left xs -> Left $ "Invalid bytestring[" <> T.pack (show bs) <> "]: expected a double but returned " <> T.pack (show xs)
+    fromPersistValue x = Left $ "PersistField Rational:Expected Rational, received: " <> T.pack (show x)
 
 instance PersistField Bool where
     toPersistValue = PersistBool
@@ -268,10 +269,10 @@ instance PersistField a => PersistField (Maybe a) where
     toPersistValue Nothing = PersistNull
     toPersistValue (Just a) = toPersistValue a
     fromPersistValue PersistNull = Right Nothing
-    fromPersistValue x = fmap Just $ fromPersistValue x
+    fromPersistValue x = Just <$> fromPersistValue x
 
 instance PersistField a => PersistField [a] where
-    toPersistValue = PersistList . map toPersistValue
+    toPersistValue = PersistList . fmap toPersistValue
     fromPersistValue (PersistList l) = fromPersistList l
     fromPersistValue (PersistText t) = fromPersistValue (PersistByteString $ TE.encodeUtf8 t)
     fromPersistValue (PersistByteString bs)
@@ -287,13 +288,13 @@ instance PersistField a => PersistField (V.Vector a) where
                             (Right . V.fromList) . fromPersistValue
 
 instance (Ord a, PersistField a) => PersistField (S.Set a) where
-    toPersistValue = PersistList . map toPersistValue . S.toList
+    toPersistValue = PersistList . fmap toPersistValue . S.toList
     fromPersistValue (PersistList list) =
-      either Left (Right . S.fromList) $ fromPersistList list
+      S.fromList <$> fromPersistList list
     fromPersistValue (PersistText t) = fromPersistValue (PersistByteString $ TE.encodeUtf8 t)
     fromPersistValue (PersistByteString bs)
         | Just values <- A.decode' (L.fromChunks [bs]) =
-            either Left (Right . S.fromList) $ fromPersistList values
+            S.fromList <$> fromPersistList values
     fromPersistValue PersistNull = Right S.empty
     fromPersistValue x = Left $ T.pack $ "Expected PersistSet, received: " ++ show x
 
@@ -301,16 +302,16 @@ instance (PersistField a, PersistField b) => PersistField (a,b) where
     toPersistValue (x,y) = PersistList [toPersistValue x, toPersistValue y]
     fromPersistValue v =
         case fromPersistValue v of
-            Right (x:y:[])  -> (,) <$> fromPersistValue x <*> fromPersistValue y
-            Left e          -> Left e
-            _               -> Left $ T.pack $ "Expected 2 item PersistList, received: " ++ show v
+            Right [x,y]  -> (,) <$> fromPersistValue x <*> fromPersistValue y
+            Left e       -> Left e
+            _            -> Left $ T.pack $ "Expected 2 item PersistList, received: " ++ show v
 
 instance PersistField v => PersistField (IM.IntMap v) where
     toPersistValue = toPersistValue . IM.toList
-    fromPersistValue = (fmap IM.fromList) . fromPersistValue
+    fromPersistValue = fmap IM.fromList . fromPersistValue
 
 instance PersistField v => PersistField (M.Map T.Text v) where
-    toPersistValue = PersistMap . map (\(k,v) -> (k, toPersistValue v)) . M.toList
+    toPersistValue = PersistMap . fmap (second toPersistValue) . M.toList
     fromPersistValue = fromPersistMap <=< getPersistMap
 
 instance PersistField PersistValue where
@@ -333,6 +334,7 @@ fromPersistMap = foldShortLeft fromPersistValue [] where
             Left e   -> Left e
             Right v' -> go ((k,v'):acc) kvs
 
+-- | FIXME Add documentation to that.
 getPersistMap :: PersistValue -> Either T.Text [(T.Text, PersistValue)]
 getPersistMap (PersistMap kvs) = Right kvs
 getPersistMap (PersistText t)  = getPersistMap (PersistByteString $ TE.encodeUtf8 t)
