@@ -1,7 +1,8 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE TypeFamilies, FlexibleContexts #-}
+{-# LANGUAGE TypeFamilies, FlexibleContexts, ConstraintKinds #-}
 module Database.Persist.Class.PersistUnique
-    ( PersistUnique (..)
+    ( PersistUniqueRead (..)
+    , PersistUniqueWrite (..)
     , getByValue
     , insertBy
     , replaceUnique
@@ -35,7 +36,11 @@ import Data.Text (unpack, Text)
 -- you must manually place a unique index on a field to have a uniqueness
 -- constraint.
 --
--- Some functions in this module ('insertUnique', 'insertBy', and
+class (PersistCore backend, PersistStoreRead backend) => PersistUniqueRead backend where
+    -- | Get a record by unique key, if available. Returns also the identifier.
+    getBy :: (MonadIO m, PersistRecordBackend record backend) => Unique record -> ReaderT backend m (Maybe (Entity record))
+
+-- | Some functions in this module ('insertUnique', 'insertBy', and
 -- 'replaceUnique') first query the unique indexes to check for
 -- conflicts. You could instead optimistically attempt to perform the
 -- operation (e.g. 'replace' instead of 'replaceUnique'). However,
@@ -44,17 +49,15 @@ import Data.Text (unpack, Text)
 --  determing the column of failure;
 --
 --  * an exception will automatically abort the current SQL transaction.
-class PersistStore backend => PersistUnique backend where
-    -- | Get a record by unique key, if available. Returns also the identifier.
-    getBy :: (MonadIO m, backend ~ PersistEntityBackend val, PersistEntity val) => Unique val -> ReaderT backend m (Maybe (Entity val))
+class (PersistUniqueRead backend, PersistStoreWrite backend) => PersistUniqueWrite backend where
 
     -- | Delete a specific record by unique key. Does nothing if no record
     -- matches.
-    deleteBy :: (MonadIO m, PersistEntityBackend val ~ backend, PersistEntity val) => Unique val -> ReaderT backend m ()
+    deleteBy :: (MonadIO m, PersistRecordBackend record backend) => Unique record -> ReaderT backend m ()
 
     -- | Like 'insert', but returns 'Nothing' when the record
     -- couldn't be inserted because of a uniqueness constraint.
-    insertUnique :: (MonadIO m, PersistEntityBackend val ~ backend, PersistEntity val) => val -> ReaderT backend m (Maybe (Key val))
+    insertUnique :: (MonadIO m, PersistRecordBackend record backend) => record -> ReaderT backend m (Maybe (Key record))
     insertUnique datum = do
         conflict <- checkUnique datum
         case conflict of
@@ -67,16 +70,32 @@ class PersistStore backend => PersistUnique backend where
     -- * update the existing record that matches the uniqueness contraint.
     --
     -- Throws an exception if there is more than 1 uniqueness contraint.
-    upsert :: (MonadIO m, PersistEntityBackend val ~ backend, PersistEntity val)
-           => val          -- ^ new record to insert
-           -> [Update val]
+    upsert :: (MonadIO m, PersistRecordBackend record backend)
+           => record          -- ^ new record to insert
+           -> [Update record]
            -- ^ updates to perform if the record already exists (leaving
            -- this empty is the equivalent of performing a 'repsert' on a
            -- unique key)
-           -> ReaderT backend m (Entity val)
+           -> ReaderT backend m (Entity record)
            -- ^ the record in the database after the operation
     upsert record updates = do
         uniqueKey <- onlyUnique record
+        upsertBy uniqueKey record updates
+
+    -- | Update based on a given uniqueness constraint or insert:
+    --
+    -- * insert the new record if it does not exist;
+    -- * update the existing record that matches the given uniqueness contraint.
+    upsertBy :: (MonadIO m, PersistRecordBackend record backend)
+            => Unique record -- ^ uniqueness constraint to find by
+            -> record          -- ^ new record to insert
+            -> [Update record]
+            -- ^ updates to perform if the record already exists (leaving
+            -- this empty is the equivalent of performing a 'repsert' on a
+            -- unique key)
+            -> ReaderT backend m (Entity record)
+            -- ^ the record in the database after the operation
+    upsertBy uniqueKey record updates = do
         mExists <- getBy uniqueKey
         k <- case mExists of
             Just (Entity k _) -> do
@@ -89,22 +108,33 @@ class PersistStore backend => PersistUnique backend where
 -- | Insert a value, checking for conflicts with any unique constraints.  If a
 -- duplicate exists in the database, it is returned as 'Left'. Otherwise, the
 -- new 'Key is returned as 'Right'.
-insertBy :: (MonadIO m, PersistEntity val, PersistUnique backend, PersistEntityBackend val ~ backend)
-         => val -> ReaderT backend m (Either (Entity val) (Key val))
+insertBy :: (MonadIO m, PersistUniqueWrite backend, PersistRecordBackend record backend)
+         => record -> ReaderT backend m (Either (Entity record) (Key record))
 insertBy val = do
     res <- getByValue val
     case res of
       Nothing -> Right `liftM` insert val
       Just z -> return $ Left z
 
+-- | Insert a value, checking for conflicts with any unique constraints. If a
+-- duplicate exists in the database, it is left untouched. The key of the
+-- existing or new entry is returned
+insertOrGet :: (MonadIO m, PersistUniqueWrite backend, PersistRecordBackend record backend)
+            => record -> ReaderT backend m (Key record)
+insertOrGet val = do
+    res <- getByValue val
+    case res of
+        Nothing -> insert val
+        Just (Entity key _) -> return key
+
 -- | Return the single unique key for a record.
-onlyUnique :: (MonadIO m, PersistEntity val, PersistUnique backend, PersistEntityBackend val ~ backend)
-           => val -> ReaderT backend m (Unique val)
+onlyUnique :: (MonadIO m, PersistUniqueWrite backend, PersistRecordBackend record backend)
+           => record -> ReaderT backend m (Unique record)
 onlyUnique record = case onlyUniqueEither record of
     Right u -> return u
     Left us -> requireUniques record us >>= liftIO . throwIO . OnlyUniqueException . show . length
 
-onlyUniqueEither :: (PersistEntity val) => val -> Either [Unique val] (Unique val)
+onlyUniqueEither :: (PersistEntity record) => record -> Either [Unique record] (Unique record)
 onlyUniqueEither record = case persistUniqueKeys record of
     [u] -> Right u
     us  -> Left us
@@ -113,7 +143,7 @@ onlyUniqueEither record = case persistUniqueKeys record of
 -- of a 'Unique' record. Returns a record matching /one/ of the unique keys. This
 -- function makes the most sense on entities with a single 'Unique'
 -- constructor.
-getByValue :: (MonadIO m, PersistEntity record, PersistUnique backend, PersistEntityBackend record ~ backend)
+getByValue :: (MonadIO m, PersistUniqueRead backend, PersistRecordBackend record backend)
            => record -> ReaderT backend m (Maybe (Entity record))
 getByValue record = checkUniques =<< requireUniques record (persistUniqueKeys record)
   where
@@ -142,7 +172,7 @@ recordName = unHaskellName . entityHaskell . entityDef . Just
 -- If uniqueness is violated, return a 'Just' with the 'Unique' violation
 --
 -- Since 1.2.2.0
-replaceUnique :: (MonadIO m, Eq record, Eq (Unique record), PersistEntityBackend record ~ backend, PersistEntity record, PersistUnique backend)
+replaceUnique :: (MonadIO m, Eq record, Eq (Unique record), PersistRecordBackend record backend, PersistUniqueWrite backend)
               => Key record -> record -> ReaderT backend m (Maybe (Unique record))
 replaceUnique key datumNew = getJust key >>= replaceOriginal
   where
@@ -161,11 +191,11 @@ replaceUnique key datumNew = getJust key >>= replaceOriginal
 --
 -- Returns 'Nothing' if the entity would be unique, and could thus safely be inserted.
 -- on a conflict returns the conflicting key
-checkUnique :: (MonadIO m, PersistEntityBackend record ~ backend, PersistEntity record, PersistUnique backend)
+checkUnique :: (MonadIO m, PersistRecordBackend record backend, PersistUniqueRead backend)
             => record -> ReaderT backend m (Maybe (Unique record))
 checkUnique = checkUniqueKeys . persistUniqueKeys
 
-checkUniqueKeys :: (MonadIO m, PersistEntity record, PersistUnique backend, PersistEntityBackend record ~ backend)
+checkUniqueKeys :: (MonadIO m, PersistEntity record, PersistUniqueRead backend, PersistRecordBackend record backend)
                 => [Unique record] -> ReaderT backend m (Maybe (Unique record))
 checkUniqueKeys [] = return Nothing
 checkUniqueKeys (x:xs) = do
